@@ -1,31 +1,98 @@
 /* Copyright (c) 1997-1998 Karl M. Hegbloom <karlheg@debian.org>
+ * (note: above email address is defunct)
  *
  *  Originally Based on `tmpwatch-1.2/1.4' RHS, Erik Troan <ewt@redhat.com>
  *
  * This program may be freely redistributed under the terms of the GNU
  * Public License.  You should be able to find a copy of the GPL in
  * your "/usr/doc/copyright" directory on most GNU/Linux installations.
+ *
+ * Now maintained by ps - Paul Slootman <tmpreaper@packages.debian.org>
+ * 2001-12-03 1.4.15
+ * Patches from * mb - Marcus Brinkmann <Marcus.Brinkmann@ruhr-uni-bochum.de>
+ * for using get_current_dir_name() and portability on HURD. ps hacked this
+ * so that it should (still?) compile on non-glibc systems.
  */
 
-#include <dirent.h>
-#include <errno.h>
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
+#define _GNU_SOURCE	/* For get_current_dir_name()  */
+
+#include <sys/types.h>
+#include <sys/stat.h>
+#if HAVE_DIRENT_H
+# include <dirent.h>
+# define NAMLEN(dirent) strlen((dirent)->d_name)
+#else
+# define dirent direct
+# define NAMLEN(dirent) (dirent)->d_namlen
+# if HAVE_SYS_NDIR_H
+#  include <sys/ndir.h>
+# endif
+# if HAVE_SYS_DIR_H
+#  include <sys/dir.h>
+# endif
+# if HAVE_NDIR_H
+#  include <ndir.h>
+# endif
+#endif
+#ifdef HAVE_ERRNO_H
+# include <errno.h>
+#endif
 #include <getopt.h>
 #include <glob.h>
-#include <linux/limits.h>
-#include <malloc.h>
+#ifdef HAVE_LIMITS_H
+# include <limits.h>
+#endif
 #include <stdarg.h>
 #include <stdio.h>
-#include <string.h>
-#include <sys/stat.h>
-#include <sys/time.h>
-#include <sys/types.h>
+#if STDC_HEADERS
+# include <string.h>
+#else
+# ifndef HAVE_MEMCPY
+#  define memmove(d, s, n) bcopy ((s), (d), (n))
+# endif
+#endif
+#ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
-#include <unistd.h>
+#endif
+#include <time.h>
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
+#ifdef HAVE_STDLIB_H
+# include <stdlib.h>
+#endif
+
+#ifdef __linux__
+# include <fcntl.h>
+# include <sys/ioctl.h>
+# include <linux/ext2_fs.h>
+# include <utime.h>
+#endif
+
+/* do we need to work around a glibc dependency? */
+#ifndef __USE_GNU
+# ifdef __linux__      /* sanity check */
+#  error "__USE_GNU not defined on a linux system, investigate! And report to tmpreaper@packages.debian.org"
+# else
+#  define NO_get_current_dir_name
+   /* XXX TODO: a workaround for missing get_current_dir_name() */
+# endif
+#endif
 
 #ifndef MAX_FORKS
 # define MAX_FORKS	200
 #endif
 
+#ifdef GLOB_BRACE
+# define TMPREAPER_GLOB_BRACE GLOB_BRACE
+#else
+# define TMPREAPER_GLOB_BRACE 0
+#endif
+
 /*
  * tmpreaper.c -- remove files in a directory, but do it carefully
  */
@@ -238,8 +305,8 @@ cleanupDirectory (const char	   * dirname ,
 
     if (FLAGS_TEST_P (flags)) {
 	message (LOG_VERBOSE,
-		 "Pretending to clean up directory `%s'.\n",
-		 dirname);
+		 "(PID %u) Pretending to clean up directory `%s'.\n",
+		 (u_int) getpid(), dirname);
     }
     else {
 	message(LOG_DEBUG,
@@ -247,16 +314,25 @@ cleanupDirectory (const char	   * dirname ,
 		dirname);
     }
     if (fork_count > MAX_FORKS) {
+#ifdef NO_get_current_dir_name
 	char	path[PATH_MAX], *p;
+#else
+	char	*path, *p;
+#endif
 	int	l;
 	/* the admin will want to know where the problem lies */
-	if (getcwd(path, PATH_MAX) < 0) {
+#ifdef NO_get_current_dir_name
+	if (getcwd(path, PATH_MAX) < 0)
 	    /* too long even for PATH_MAX, show something anyway */
+#else
+	if ((path = get_current_dir_name()) < 0)
+#endif
+        {
 	    p = "..../";
 	}
 	else {
 	    l = strlen(path);
-	    if (l > 60) {
+	    if (l > 65) {
 		    /* remove middle part of the path */
 		    p = path + l - 30;
 		    strcpy(&path[30], " ... ");
@@ -284,11 +360,18 @@ cleanupDirectory (const char	   * dirname ,
      * right choice by having chosen the `recursive forking' algorithm.  It is
      * much easier to implement and understand as well as being more secure
      * through avoiding `cd ..'. - karlheg
+     *
+     * The geek girl link is dead.  fork() is not efficient on all systems,
+     * so it would still be better to use a directory stack + fchdir(). - mb
+     *
+     * This is another link to what geek-girl had:
+     * http://www.geocrawler.com/mail/msg.php3?msg_id=184906&list=91
+     * - ps
      */
 
     if (! (pid = fork())) {
 	    if (safe_chdir (dirname))
-		    return 1;
+		    exit(1);
 	    fork_count++;
 	    if (! (dir = opendir ("."))) {
 		message (LOG_FATAL,
@@ -353,8 +436,8 @@ cleanupDirectory (const char	   * dirname ,
 		if (S_ISDIR (sb.st_mode)) {
 		    cleanupDirectory (ent->d_name, killTime, flags);
 		    message (LOG_VERBOSE,
-			     "Back from recursing down `%s'.\n",
-			     ent->d_name);
+			     "(PID %u) Back from recursing down `%s'.\n",
+			     (u_int) getpid(), ent->d_name);
 		}
 
 		if (FLAGS_PROTECT_P (flags)) {
@@ -414,9 +497,41 @@ cleanupDirectory (const char	   * dirname ,
 				 "Removing file `%s/%s'.\n",
 				 dirname, ent->d_name);
 			if (unlink (ent->d_name))
-			    message (LOG_ERROR,
+#if defined(__linux__)
+                        {
+                            int give_error = 1;
+                            int fd;
+                            if ((fd = open(ent->d_name, O_RDONLY)) >= 0) {
+                                unsigned int flags;
+                                if (ioctl(fd, EXT2_IOC_GETFLAGS, &flags) >= 0) {
+                                    if (flags & EXT2_IMMUTABLE_FL) {
+                                        give_error = 0;
+                                    }
+                                }
+                                close(fd);
+                                /* Restore access time! */
+                                /* However, only bother if we give a warning. */
+                                /* With no warning, it doesn't matter.        */
+                                /* And it delays the next attempt :-)         */
+                                if (give_error) {
+                                    struct utimbuf ts;
+                                    ts.actime  = sb.st_atime;
+                                    ts.modtime = sb.st_mtime;
+                                    utime(ent->d_name, &ts);
+                                }
+                            }
+                            if (give_error)
+#endif
+                                message (LOG_ERROR,
 				     "Failed to unlink `%s': %s\n",
 				     ent->d_name, strerror (errno));
+#if defined(__linux__)
+                            else
+                                message (LOG_VERBOSE,
+                                    "file `%s' marked immutable -- skipping.\n",
+                                    ent->d_name);
+                        }
+#endif
 		    }
 		}
 		else {
@@ -438,6 +553,11 @@ cleanupDirectory (const char	   * dirname ,
 	    exit (0);
     }
 
+    if (pid < 0) {
+        message(LOG_ERROR, "fork failed: %s\n", strerror(errno));
+        return 1;
+    }
+
     if (waitpid(pid, &status, 0) < 0) {
 	message(LOG_ERROR, "waitpid failed: %s\n", strerror(errno));
 	return 1;
@@ -453,8 +573,12 @@ void
 printCopyright (void)
 {
     fprintf (stderr,
+#ifdef DEBIAN
+	     "tmpreaper -- Version: " VERSION "-DEB\n"
+#else
 	     "tmpreaper -- Version: " VERSION "\n"
-	     "(c) 1997 Karl M. Hegbloom <karlheg@debian.org>\n"
+#endif
+	     "(c) 1997 Karl M. Hegbloom <tmpreaper@packages.debian.org>\n"
 	     "This may be freely redistributed under the terms of the GNU Public License.\n");
 }
 /*	               1         2         3         4         5         6         7         8
@@ -621,7 +745,12 @@ main (int     argc,
      * See: `APUE', W. Richard Stevens,  pp. 189-190
      * Bug pointed out by Joey Hess <joey@kitenet.net>
      */
-    if (setvbuf (stdout, (void *)NULL, _IOLBF, BUFSIZ) != 0) {
+#ifdef SETVBUF_REVERSED
+    if (setvbuf (stdout, _IOLBF, (void *)NULL, BUFSIZ) != 0)
+#else
+    if (setvbuf (stdout, (void *)NULL, _IOLBF, BUFSIZ) != 0)
+#endif
+    {
 	message (LOG_FATAL,
 		 "setvbuf() failed for `stdout'.\n");
 	/*NOTREACHED*/
@@ -638,12 +767,21 @@ main (int     argc,
 
     killTime = time (NULL) - grace;
 
-    if ((pwd = getcwd (NULL, 0)) == NULL) {
+    /* use getcwd() to test whether current directory is valid */
+    if (!(pwd = malloc(PATH_MAX))) {
 	message (LOG_FATAL,
-		 "getcwd() failed: %s",
+		 "malloc for getcwd failed: %s\n",
 		 strerror(errno));
 	/*NOTREACHED*/
     }
+    if (getcwd (pwd, PATH_MAX) == NULL) {
+	message (LOG_FATAL,
+		 "getcwd() failed: %s\n",
+		 strerror(errno));
+	/*NOTREACHED*/
+    }
+    /* don't need pwd, but leave it allocated for possible debug things below */
+
     /* For each directory in the cleanup list, cd there, glob, then delete */
     for (adir = optind; adir < argc; adir++) {
 	if ((ret = chdir (argv[adir])) != 0) {
@@ -652,17 +790,20 @@ main (int     argc,
 		     argv[adir], strerror(errno));
 	    continue;
 	}
-	message (LOG_DEBUG,
+	if (LOG_DEBUG >= logLevel) {
+	    /* this test is to prevent non-necessary calls to getcwd() */
+	    message (LOG_DEBUG,
 		 "getcwd() is: %s\n",
-		 getcwd (NULL, 0));
+		 getcwd (pwd, PATH_MAX));
+	}
 	/*  ... for each protect_argv, glob, using GLOB_APPEND */
 	if (FLAGS_PROTECT_P (flags)) {
 	    pp = p;
 	    while (*pp != NULL) {
 		if (p == pp) {
-		    ret = glob (*pp, (GLOB_NOSORT | GLOB_BRACE), NULL, &protect_glob);
+		    ret = glob (*pp, (GLOB_NOSORT | TMPREAPER_GLOB_BRACE), NULL, &protect_glob);
 		} else {
-		    ret = glob (*pp, (GLOB_APPEND | GLOB_NOSORT | GLOB_BRACE), NULL, &protect_glob);
+		    ret = glob (*pp, (GLOB_APPEND | GLOB_NOSORT | TMPREAPER_GLOB_BRACE), NULL, &protect_glob);
 		}
 		if (ret == GLOB_NOSPACE) {
 		    message (LOG_FATAL,
@@ -670,9 +811,12 @@ main (int     argc,
 		    /*NOTREACHED*/
 		}
 		if (ret == GLOB_NOMATCH) {
-		    message (LOG_VERBOSE,
+		    if (LOG_VERBOSE >= logLevel) {
+			/* test to prevent non-necessary calls to getcwd() */
+			message (LOG_VERBOSE,
 			     "Nothing in `%s' matches `%s'.\n",
-			     getcwd (NULL, 0), *pp);
+			     getcwd (pwd, PATH_MAX), *pp);
+		    }
 		}
 		pp++;
 	    }
